import java.io.*;
import java.awt.event.*;
import java.awt.*;

/**
	Bei der Klasse <tt>ToyEmulatorKernel</tt> handelt es sich um die 
	eigentliche Hauptklasse des Toy-Programmpaketes. Hier wird der
	Akkumulator, der Programmzhler sowie das Instruktionsregister
	verwaltet. Auch auf Speicherinhalte wird von hier aus zugegriffen. 
	Ausserdem existieren Referenzen zu fast allen weiteren Klassen des 
	Programmpaketes. Alle anzuzeigenden Informationen werden
	an die Klasse <tt>ToyEmulatorGUI</tt> weitergeleitet. 
*/
public class ToyEmulatorKernel extends Object {

	/**
		Hier wird eine neue GUI-Instanz angelegt, sowie zwei
		Initialisierungsmethoden aufgerufen.
	*/ 
	public ToyEmulatorKernel() {
	
		//	Anlegen einer neuen GUI-Instanz 
		TEGUI = new ToyEmulatorGUI(this);
		//	Setzen aller Anfangswerte 
		initializeToyEmulatorKernel();
		initializeToyEmulatorGUI();
	}


	/**
		Einlesen eines neuen Quelltextes aus einer Datei in einen
		Stringbuffer. Dieser wird dann in der Methode
		<tt>loadmemory</tt> in den Speicher eingelesen.
		
		@param file: Zeichenkette, die den Namen der zu ladenden Datei
		enthlt.
	*/
	public void loadFile(String file) {

		FileReader f;
		int c;
		try {

			currentFileName = file;
			
			//	Zunchst wird die Datei geffnet
			f = new FileReader(file);
			mem = null;
			mem = new StringBuffer();

			//	nun wird der Dateiinhalt in eine Zeichenkette geschrieben
			while ((c = f.read()) != -1) {
				mem.append((char)c);
			}
			f.close();
		} catch (IOException exc) {
			System.out.println("Fehler beim Lesen der Datei");
		}

		//	Diese Zeichenkette wird dann, entsprechend bearbeitet, in den Speicher geladen
		loadMemory(mem.toString());	
	}


	/**
		Diese Methode bewirkt ein Neuladen der aktuell geladenen
		Quelltextdatei sowie den Reset des Programmzhlers auf die erste
		Instrktion. Auch der Speicher wird in den ursprnglichen Zustand
		zurckversetzt.
	*/
	public void reset() {
			
		loadFile(currentFileName);
	}


	/**
		Diese Methode initialisiert alle Variablenwerte der Klasse auf
		zulssige Anfangswerte.
	*/
	private void initializeToyEmulatorKernel() {

		try {
			this.accumulator = new WholeHexNumber("0");	//	Akku = 0h
			this.pc = new WholeHexNumber("0");		//	ProgramCounter = 0h
			this.insnreg = new WholeHexNumber("0");		//	Befehlsreg = 0h
		} catch (NotAHexNumberException exc) {
			System.out.println(exc.getMessage());
		}
		this.mnemonic = "";														//	Mnemonic = ""
		if(memory == null) this.memory = new ToyMemory();						//	Speicher initialisieren
		else memory.resetMemory();
	}


	/**
		Diese Methode initialisiert alle Variablenwerte der Klasse
		<tt>ToyEmulatorGUI</tt> auf zulssige Anfangswerte.
	*/
	private void initializeToyEmulatorGUI() {

		TEGUI.setAccuTextField(accumulator.toString());
		TEGUI.setAccuDecTextField(Integer.toString(accumulator.toInteger()));
		TEGUI.setPCTextField(pc.toString());
		TEGUI.setInsnregTextField(insnreg.toString());
		TEGUI.setMnemonicTextField(mnemonic);
		TEGUI.setStepTextField(Integer.toString(stepWidth));
		TEGUI.setMemoryTextArea(memory.toString(true));
		TEGUI.setMemoryDataTextArea(memory.toString(false));	
	}


	/**
		Einlesen des Quelltextes eines Toy-Programmes in den Speicher
		sowie anschlieende Aktualisierung aller Toy-Parameter.
		@param s: Zeichenkette, die den Quelltext enthlt 
	*/	
	private void loadMemory(String s) {
		BufferedReader f;
		int 	linecounter = 0, linenr, equalspos;
		String line;
		try {
			f = new BufferedReader(new StringReader(s));
			initializeToyEmulatorKernel();
			while ((line = f.readLine()) != null) {
				if (line.length() != 0) {   			// keine leere Zeile
					if(line.charAt(0) == '$') {		// keine Kommentarzeile
						memory.setCell(linecounter,line.substring(1,5));
						linecounter++;
					} else if(line.charAt(0) == '%') {
						equalspos = line.indexOf("=");
						linenr = Integer.parseInt(line.substring(1,equalspos),16);
						memory.setCell(linenr,line.substring(equalspos+1,equalspos+5));
					}
				}
			}
			//	Jetzt alles aktualisieren.
			TEGUI.setAccuTextField(accumulator.toString());
			TEGUI.setAccuDecTextField(Integer.toString(accumulator.toInteger()));
			TEGUI.setInsnregTextField(memory.getCell(pc.toInteger()).toString());
			insnreg = getInsnregFromMemory();
			TEGUI.setMnemonicTextField(getMnemonicFromInsnreg());
			TEGUI.setMemoryTextArea(memory.toString(true));
			TEGUI.setMemoryDataTextArea(memory.toString(false));
			TEGUI.setPCTextField(pc.toString());
			f.close();

			// Anzeigen aller Vernderungen in archGUI
			if(archGUI!=null) {
				archGUI.setDoNextStep(true);
				archGUI.setArchTextFields();
				archGUI.setDoNextStep(false);
			}
		} catch (IOException exc) {
			System.out.println("Fehler in Datei");
		}
	}


	/** 
		Ausrechnen des Accu-Wertes, schon bevor dies tatschlich getan
		wird. Wird fr Darstellung im Architekturfenster bentigt.
		@return Akkuwert, der nach der nchsten Instruktion entsteht
	*/
	public String preCalcAccuValue()
	{
		int instr_int = getInsnregFromMemory().toInteger();
		int arg_mem = (getArgumentFromMemory(pc.toInteger())).toInteger();
		
		try {		
			WholeHexNumber whn = new WholeHexNumber(memory.getCell(arg_mem));

			switch (instr_int) {
				case 1:		return (memory.getCell(arg_mem)).toString();
				case 3:		return (WholeHexNumber.HexADDHex(whn,accumulator)).toString();
				case 4:		return (WholeHexNumber.HexSUBHex(whn,accumulator)).toString();
				case 5:		return (WholeHexNumber.HexORHex(whn,accumulator)).toString();
				case 6:		return (WholeHexNumber.HexANDHex(whn,accumulator)).toString(); 
				case 7:		return (WholeHexNumber.HexXORHex(whn,accumulator)).toString(); 
				case 8:		return (WholeHexNumber.NOTHex(accumulator)).toString(); 
				case 9:		return (WholeHexNumber.INCHex(accumulator)).toString(); 
				case 10:	return (WholeHexNumber.DECHex(accumulator)).toString(); 
				case 11:	return "0000";
			}		
		} catch (Exception e) {}
		return accumulator.toString();
	}


	/**
		Methode zum aktualisieren des privaten Attributes
		<tt>accumulator</tt>.
		@param newAccuValue: Neuer Akkumulatorwert
	*/ 
	public void setAccuValue(String newAccuValue) {
	
		try {
			accumulator = new WholeHexNumber(newAccuValue);
		} catch (NotAHexNumberException exc) {
			System.out.println(exc.getMessage());
		}
	}


	/**
		Methode um den aktuellen Akkumulatorwert zu ermitteln.
		@return Akkumulatorwert in Stringreprsentation
	*/
	public String getAccuValue() {
		return accumulator.toString();
	}
	
	
	/**
		Methode zum Aktualisieren des privaten Attributes <tt>pc</tt>.
		@param newPCValue: Neuer Programmzhlerwert
	*/ 
	public void setPCValue(String newPCValue) {

		try {
			pc = new WholeHexNumber(newPCValue);
		} catch (NotAHexNumberException exc) {
			System.out.println(exc.getMessage());
		}
	}

	/**
		Gibt den aktuellen Wert des Programmzhlers als String zurck
		@return pc als string
	*/	
	public String getPCValue() {
		return pc.toString();
	}


	/**
		Setzt die Speicherstelle Nummer <tt>adr</tt> auf den Wert
		<tt>value</tt>.
		@param adr: Adresse der zu ndernden Speicherstelle
		@param value: Neuer Wert der zu ndernden Speicherstelle
	*/
	public void setMemoryValue(int adr, String value) {

		memory.setCell(adr,value);
		TEGUI.setMemoryDataTextArea(memory.toString(false));
	}	

	
	/**
		Methode zum Aktualisieren des privaten Attributes <tt>gomode</tt>.
		@param mode: Neuer gomode-wert
	*/
	public void setgomode(boolean mode) {

		gomode = mode;
	}

	
	/**
		Gibt den aktuellen Wert des Attributes <tt>gomode<tt> zurck
		@return <tt>Wert von gomode</tt>
	*/
	public boolean getgomode() {

		return gomode;
	}

	/**
		Methode zum Aktualisieren des privaten Attributes <tt>stepWidth</tt>.
		@param width: Neuer stepWidth-wert
	*/
	public void setStepWidth(int width) {
		
		stepWidth = width;
	}
	
	
	/**
		Gibt den aktuellen Wert des Attributes <tt>stepWidth<tt> zurck
		@return <tt>Wert von stepWidth</tt>
	*/
	public int getStepWidth() {
		
		return stepWidth;
	}		


	/**	
		Diese Methode liefert den Instructioncode aus der aktuellen
		Speicherzelle.
		@return aktuelle Instruktion als Hexadezimalnummer
	*/
	public WholeHexNumber getInsnregFromMemory() {

		int pc_int = pc.toInteger();
		WholeHexNumber insn_whn = null;
		try {
			//	Der Instructioncode steht in in der ersten HexZiffer
			insn_whn = new WholeHexNumber((memory.getCell(pc_int)).substring(0,1));
		} catch (NotAHexNumberException exc) {
			System.out.println(exc.getMessage());
		}
		return (insn_whn);
	}


	/**
		Die Methode liefert die Mnemonic zum aktuellen Instructioncode.
		@return Mnemonic als String
	*/
	private String getMnemonicFromInsnreg() {
		
		return getMnemonicFromInt(insnreg.toInteger());
	}
	
	
	/**
		Die Methode liefert die Mnemonic zum bergebenem Integerwert.
		@param i: Instruktionscode
		@return Mnemonic als String
	*/
	public static String getMnemonicFromInt(int i) {

		String m = "";
		switch (i) {
			case 0:		m = "STO"; break;
			case 1:		m = "LDA"; break;
			case 2:		m = "BRZ"; break;
			case 3:		m = "ADD"; break;
			case 4:		m = "SUB"; break;
			case 5:		m = "OR"; break;
			case 6:		m = "AND"; break;
			case 7:		m = "XOR"; break;
			case 8:		m = "NOT"; break;
			case 9:		m = "INC"; break;
			case 10:	m = "DEC"; break;
			case 11:	m = "ZRO"; break;
			case 12:	m = "NOP"; break;
			case 13:	m = "NOP"; break;				
			case 14:	m = "NOP"; break;				
			case 15:	m = "STOP"; break;				
		}
		return m;
	}
	

	/**
		Liefert Argument (Inhalt ohne erstes Zeichen) der an Nummer
		<tt>adr</tt> gelegenen Speicheradresse.
		@param adr: Adresse der gesuchten Speicheradresse
		@return Argument als Hexadezimalzahl
	*/
	public WholeHexNumber getArgumentFromMemory(int adr) {

		WholeHexNumber arg = null;
		try {
			arg = new WholeHexNumber((memory.getCell(adr)).substring(1,4));
		} catch (NotAHexNumberException exc) {
			System.out.println(exc);
		} catch (StringIndexOutOfBoundsException exc1) {
			try { arg = new WholeHexNumber("0"); }
			catch (NotAHexNumberException exc2) {
				System.out.println(exc2);
			}
		}
		return arg;
	}


	/**
		Gibt den aktuellen Wert des Attributes <tt>memory<tt> zurck
		@return <tt>Wert von memory</tt>
	*/
	public ToyMemory getToyMemory()
	{
	    return memory;
	}
	
	
	/**
		Gibt den aktuellen Wert des Attributes <tt>archGUI<tt> zurck
		@return <tt>Wert von archGUI</tt>
	*/
	public TArchGUI getArchGUI()
	{
		return archGUI;
	}


	/**
		Mit der step()-Methode wird in der Quelltextabarbeitung ein,
		oder eine von <tt>stepWidth</tt> vorgegebene Anzahl von
		Schritten,  fortgefahren. Abhngig vom Instructioncode wird dann
		eine der mglichen Anweisungen ausgefhrt und anschlieend alle
		Vernderungen sichtbar gemacht */
	public void step() {

		int i, instr_int = getInsnregFromMemory().toInteger();
		boolean repaintMemData = false;
		
		
		for(i=0;(i<stepWidth)&&(instr_int!=15);i++) {
			switch (instr_int) {
				case 0:		STORE((getArgumentFromMemory(pc.toInteger())).toInteger()); repaintMemData=true; break;
				case 1:		LOAD((getArgumentFromMemory(pc.toInteger())).toInteger()); break;
				case 2:		JMPZ(getArgumentFromMemory(pc.toInteger())); break;
				case 3:		ADD((getArgumentFromMemory(pc.toInteger())).toInteger()); break;
				case 4:		SUB((getArgumentFromMemory(pc.toInteger())).toInteger()); break;
				case 5:		OR((getArgumentFromMemory(pc.toInteger())).toInteger()); break;
				case 6:		AND((getArgumentFromMemory(pc.toInteger())).toInteger()); break;
				case 7:		XOR((getArgumentFromMemory(pc.toInteger())).toInteger()); break;
				case 8:		NOT(); break;
				case 9:		INC(); break;
				case 10:	DEC(); break;
				case 11:	ZERO(); break;
				case 12:	NOP(); break;
				case 13:	NOP(); break;
				case 14:	NOP(); break;
				case 15:	STOP(); break;
			}
			instr_int = getInsnregFromMemory().toInteger();
		}

		//	Anzeigen aller Vernderungen in TEGUI
		insnreg = getInsnregFromMemory();
		TEGUI.setAccuTextField(accumulator.toString());
		TEGUI.setAccuDecTextField(Integer.toString(accumulator.toInteger()));
		TEGUI.setInsnregTextField(memory.getCell(pc.toInteger()).toString());
		TEGUI.setMnemonicTextField(getMnemonicFromInsnreg());
		if(repaintMemData) TEGUI.setMemoryDataTextArea(memory.toString(false));
		TEGUI.setPCTextField(pc.toString());
		
		// Anzeigen aller Vernderungen in archGUI
		if(archGUI!=null) {
			archGUI.setDoNextStep(true);
			archGUI.setArchTextFields();
			archGUI.setDoNextStep(false);
			archGUI.repaint();
		}
	}


	/**
		Alle Instructions werden nacheinander ausgefhrt, bis entweder
		im Quelltext die STOP-Instruktion auftaucht, oder eine beliebige
		Maustaste gedrckt wird. Zum Ausfhren der Instruktionen wird
		die <tt>step</tt>-Methode aufgerufen.
	*/
	public void go() {

		EventQueue eq = Toolkit.getDefaultToolkit().getSystemEventQueue();
		AWTEvent evt;
		ActionEvent aevt;
			
		int instr_int = getInsnregFromMemory().toInteger();
		while ((instr_int != 15) && (gomode)) {
			step();
			instr_int = getInsnregFromMemory().toInteger();
			try {
				while(eq.peekEvent() != null) {
					evt = eq.getNextEvent();
					if (evt.getID() == MouseEvent.MOUSE_CLICKED) {
						gomode=false;
						eq.postEvent(evt);
						break;
					}
				}
			} catch(InterruptedException e) {} 
		}
		TEGUI.getMessageLabel().setText("");
		TEGUI.getGoButton().setLabel("GO");
	}


	/**
		Diese Methode erstellt oder zerstrt das Architekturfenster, je
		nachdem ob es zur Zeit vorhanden ist oder nicht. Das Attribut
		<tt>archGUI</tt> wird dafr verwendet.
	*/
	public void arch() {
		if(archGUI != null) {
			TEGUI.setArchButtonLabel("Show Architecture");
			archGUI.dispose();
			archGUI = null;
			
		} else {
			TEGUI.setArchButtonLabel("Hide Architecture");
			archGUI = new TArchGUI(this);
			archGUI.show();
			
		}
	}
		
	
	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>STORE</tt>
		@param adr: Zieladresse des zu speichernden Wertes
	*/
	private void STORE(int adr) {

		memory.setCell(adr,accumulator.toString());
		pc = WholeHexNumber.INCHex(pc);
	} 
			

	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>LOAD</tt>
		@param adr: Zieladresse des zu ladenden Wertes
	*/
	private void LOAD(int adr) {

		setAccuValue(memory.getCell(adr));
		pc = WholeHexNumber.INCHex(pc);
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>JMPZ</tt>
		@param adr: Vergleichsadresse
	*/
	private void JMPZ(WholeHexNumber adr) {

		if (accumulator.toInteger() == 0) pc = adr;
		else pc = WholeHexNumber.INCHex(pc);
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>ADD</tt>
		@param adr: Adresse des Wertes, der zum Accu addiert wird
	*/
	private void ADD(int adr) {

		try {
			WholeHexNumber whn = new WholeHexNumber(memory.getCell(adr));
			accumulator = WholeHexNumber.HexADDHex(whn,accumulator);
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}

	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>SUB</tt>
		@param adr: Adresse des Wertes, der vom Accu subtrahiert wird
	*/
	private void SUB(int adr) {

		try {
			WholeHexNumber whn = new WholeHexNumber(memory.getCell(adr));
			accumulator = WholeHexNumber.HexSUBHex(accumulator,whn);
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>OR</tt>
		@param adr: Vergleichsadresse
	*/
	private void OR(int adr) {

		try {
			WholeHexNumber whn = new WholeHexNumber(memory.getCell(adr));
			accumulator = WholeHexNumber.HexORHex(whn,accumulator);
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>AND</tt>
		@param adr: Vergleichsadresse
	*/
	private void AND(int adr) {

		try {
			WholeHexNumber whn = new WholeHexNumber(memory.getCell(adr));
			accumulator = WholeHexNumber.HexANDHex(whn,accumulator);
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>XOR</tt>
		@param adr: Vergleichsadresse
	*/
	private void XOR(int adr) {

		try {
			WholeHexNumber whn = new WholeHexNumber(memory.getCell(adr));
			accumulator = WholeHexNumber.HexXORHex(whn,accumulator);
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>NOT</tt>
	*/
	private void NOT() {

		try {
			WholeHexNumber whn = new WholeHexNumber("0");
			accumulator = WholeHexNumber.NOTHex(accumulator);
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>INC</tt>
	*/
	private void INC() {

		accumulator = WholeHexNumber.INCHex(accumulator);
		pc = WholeHexNumber.INCHex(pc);
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>DEC</tt>
	*/
	private void DEC() {

		accumulator = WholeHexNumber.DECHex(accumulator);
		pc = WholeHexNumber.INCHex(pc);
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>ZERO</tt>
		@param adr: Vergleichsadresse
	*/
	private void ZERO() {

		try {
			accumulator = new WholeHexNumber("0");
			pc = WholeHexNumber.INCHex(pc);
		} catch (Exception e) {
			;
		}
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>NOP</tt>
		@param adr: Vergleichsadresse
	*/
	private void NOP() {

		pc = WholeHexNumber.INCHex(pc);
	}


	/**
		Methode zum korrekten Abarbeiten der Toy-Anweisung
		<tt>STOP</tt>
	*/
	private void STOP() {

		;
	}
	
	private WholeHexNumber accumulator = null;			//	Akku
	private WholeHexNumber pc = null;						//	Program Counter
	private WholeHexNumber insnreg = null;				//	Instruction Register
	private String mnemonic;											//	Mnemonic
	private ToyEmulatorGUI TEGUI;							//	zugehriges Grafikinterface
	private StringBuffer mem = null;						//	Zwischenspeicher fr Speicherop.
	private ToyMemory memory = null;						//	Speicher des Toy-Rechners
	private TArchGUI archGUI = null;						// optionales Architekturfenster
	private boolean gomode = false;							// sind wir im gomode?
	private String currentFileName = "";					// zur Zeit geladenes File
	private int stepWidth = 1;								// wieviele Schritte macht ein Step

}